'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var jsCore = require('@lorisleiva/js-core');
var jsWeb3jsAdapters = require('@lorisleiva/js-web3js-adapters');
var web3_js = require('@solana/web3.js');

const ACCOUNT_HEADER_SIZE = 128n;
class Web3JsRpc {
  constructor(context, endpoint, rpcOptions) {
    this.context = context;
    this.connection = new web3_js.Connection(endpoint, rpcOptions);
    this.cluster = jsCore.resolveClusterFromEndpoint(endpoint);
  }
  getEndpoint() {
    return this.connection.rpcEndpoint;
  }
  getCluster() {
    return this.cluster;
  }
  async getAccount(address, options = {}) {
    const account = await this.connection.getAccountInfo(jsWeb3jsAdapters.toWeb3JsPublicKey(address), options);
    return this.parseMaybeAccount(account, address);
  }
  async getAccounts(addresses, options = {}) {
    const accounts = await this.connection.getMultipleAccountsInfo(addresses.map(jsWeb3jsAdapters.toWeb3JsPublicKey), options);
    return accounts.map((account, index) => this.parseMaybeAccount(account, addresses[index]));
  }
  async getProgramAccounts(programId, options = {}) {
    const accounts = await this.connection.getProgramAccounts(jsWeb3jsAdapters.toWeb3JsPublicKey(programId), {
      ...options,
      filters: options.filters?.map(filter => this.parseDataFilter(filter))
    });
    return accounts.map(({
      pubkey,
      account
    }) => this.parseAccount(account, jsWeb3jsAdapters.fromWeb3JsPublicKey(pubkey)));
  }
  async getBalance(address, options = {}) {
    const balanceInLamports = await this.connection.getBalance(jsWeb3jsAdapters.toWeb3JsPublicKey(address), options);
    return jsCore.lamports(balanceInLamports);
  }
  async getRent(bytes, options = {}) {
    const rentFor = bytes => this.connection.getMinimumBalanceForRentExemption(bytes, options.commitment);
    if (options.includesHeaderBytes ?? false) {
      const headerRent = await rentFor(0);
      const rentPerByte = ACCOUNT_HEADER_SIZE / BigInt(headerRent);
      return jsCore.lamports(rentPerByte * BigInt(bytes));
    }
    return jsCore.lamports(await rentFor(bytes));
  }
  async getLatestBlockhash(options = {}) {
    return this.connection.getLatestBlockhash(options);
  }
  async accountExists(address, options = {}) {
    return !jsCore.isZeroAmount(await this.getBalance(address, options));
  }
  async airdrop(address, amount, options = {}) {
    const signature = await this.connection.requestAirdrop(jsWeb3jsAdapters.toWeb3JsPublicKey(address), Number(amount.basisPoints));
    if (options.strategy) {
      await this.confirmTransaction(jsCore.base58.serialize(signature), options);
      return;
    }
    await this.confirmTransaction(jsCore.base58.serialize(signature), {
      ...options,
      strategy: {
        type: 'blockhash',
        ...(await this.getLatestBlockhash())
      }
    });
  }
  async call(method, params, options = {}) {
    const client = this.connection._rpcClient;
    return new Promise((resolve, reject) => {
      const callback = (error, result) => error ? reject(error) : resolve(result);
      client.request(method, params, options.id ?? null, callback);
    });
  }
  async sendTransaction(transaction, options = {}) {
    try {
      const signature = await this.connection.sendRawTransaction(this.context.transactions.serialize(transaction), options);
      return jsCore.base58.serialize(signature);
    } catch (error) {
      let resolvedError = null;
      if (error instanceof Error && 'logs' in error) {
        resolvedError = this.context.programs.resolveError(error, transaction);
      }
      throw resolvedError || error;
    }
  }
  async confirmTransaction(signature, options) {
    return this.connection.confirmTransaction(this.parseConfirmStrategy(signature, options), options.commitment);
  }
  parseAccount(account, address) {
    return {
      ...account,
      owner: jsWeb3jsAdapters.fromWeb3JsPublicKey(account.owner),
      address,
      lamports: jsCore.lamports(account.lamports)
    };
  }
  parseMaybeAccount(account, address) {
    return account ? {
      ...this.parseAccount(account, address),
      exists: true
    } : {
      exists: false,
      address
    };
  }
  parseDataFilter(filter) {
    if (!('memcmp' in filter)) return filter;
    const {
      bytes,
      ...rest
    } = filter.memcmp;
    return {
      memcmp: {
        ...rest,
        bytes: jsCore.base58.deserialize(bytes)[0]
      }
    };
  }
  parseConfirmStrategy(signature, options) {
    if (options.strategy.type === 'blockhash') {
      return {
        ...options.strategy,
        signature: jsCore.base58.deserialize(signature)[0]
      };
    }
    return {
      ...options.strategy,
      signature: jsCore.base58.deserialize(signature)[0],
      nonceAccountPubkey: jsWeb3jsAdapters.toWeb3JsPublicKey(options.strategy.nonceAccountPubkey)
    };
  }
}

exports.ACCOUNT_HEADER_SIZE = ACCOUNT_HEADER_SIZE;
exports.Web3JsRpc = Web3JsRpc;
//# sourceMappingURL=Web3JsRpc.cjs.map
